# -*- coding: utf-8 -*-

import typing as T
import sys
import json

import attr
from attrs_mate import AttrsClass

from .script_filter_object import ScriptFilterObject
from .item import Item


@attr.define
class ScriptFilter(ScriptFilterObject):
    """
    The script filter return object that flush out to standard out.

    :param items: list of items object that for drop down menu
    :param variables: session level variables. items also has item level
        variables, this feature won't be used for 99% of the time
    :param rerun: 0.1 ~ 5.0 if available, the script will only be re-run
        if the script filter is still active and the user hasn't changed
        the state of the filter by typing and triggering a re-run.

    Ref:

    - https://www.alfredapp.com/help/workflows/inputs/script-filter/json/
    """
    items: T.List[Item] = Item.ib_list_of_nested()
    variables: dict = AttrsClass.ib_dict(default=None)
    rerun: float = AttrsClass.ib_float(default=None)

    def to_script_filter(self) -> dict:
        dct = super(ScriptFilter, self).to_script_filter()
        if "items" not in dct:
            dct["items"] = list()
        return dct

    def send_feedback(self):
        """
        Flush script filter object to standard output.
        """
        json.dump(self.to_script_filter(), sys.stdout, ensure_ascii=False)
        sys.stdout.flush()



# -*- coding: utf-8 -*-

"""

"""

import typing as T
import sys
import json
import traceback
from datetime import datetime

import attr
from attrs_mate import AttrsClass

from .handler import Handler
from .path import dir_lib, dir_afwf, p_last_error, p_debug_log
from .script_filter import ScriptFilter
from .item import Icon, Item
from .icon import IconFileEnum


def log_last_error():  # pragma: no cover
    """
    Log the last exception trace back info to ``~/.alfred-afwf/last-error.txt``
    file.
    """
    traceback_msg = traceback.format_exc()
    try:
        p_last_error.write_text(traceback_msg, encoding="utf-8")
    except FileNotFoundError:
        dir_afwf.mkdir(parents=True, exist_ok=True)
        p_last_error.write_text(traceback_msg, encoding="utf-8")
    except Exception as e:
        raise e


def log_debug_info(info: str):  # pragma: no cover
    """
    Call this function anywhere. It will append the ``info`` string to the end
    of ``~/.alfred-afwf/debug.txt`` file.
    """
    try:
        with p_debug_log.open("a", encoding="utf-8") as f:
            f.write(info + "\n")
    except FileNotFoundError:
        dir_afwf.mkdir(parents=True, exist_ok=True)
        with p_debug_log.open("a", encoding="utf-8") as f:
            f.write(info + "\n")
    except Exception as e:
        raise e


@attr.define
class Workflow(AttrsClass):
    """
    Represents an Alfred Workflow object. The workflow can register many handlers.

    Each handler will be responsible for a Script Filter implementation.

    [CN]

    表示一个 Alfred Workflow 对象. 一个 workflow 可以注册多个 handler. 每个 handler
    对应着一个 Script Filter 的实现.
    """
    handlers: T.Dict[str, Handler] = attr.ib(factory=dict)

    def __attrs_post_init__(self):
        if dir_lib.exists():
            sys.path.append(str(dir_lib))

    def register(self, handler: Handler):
        """
        Register a handler to the workflow.
        """
        if handler.id in self.handlers:
            raise KeyError
        else:
            self.handlers[handler.id] = handler

    def get(self, handler_id: str) -> Handler:
        """
        Get handler by id.
        """
        return self.handlers[handler_id]

    def _run(
        self,
        arg: T.Optional[str] = None,
        debug: bool =False,
    ) -> ScriptFilter:
        """
        Low level script filter runner. It locates the handler by ``handler_id``,
        and then call the handler to generate the script filter result, and then
        flush the result to stdout, hence you can see it in the drop-down menu.

        :param debug: flag to turn on debug.
        """
        if debug:
            now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            log_debug_info(f"--- run script filter at {now} ---")

        if arg is None:
            arg = sys.argv[1]  # in format of "{handler_id} {query}"

        if debug:
            log_debug_info(f"received argument is: {arg!r}")

        handler_id, query = arg.split(" ", 1)

        if debug:
            log_debug_info(f"received handler_id is: {handler_id!r}")
            log_debug_info(f"received query is: {query!r}")

        handler = self.get(handler_id)
        sf = handler.handler(query)
        json.dump(sf.to_script_filter(), sys.stdout)
        sys.stdout.flush()
        return sf

    def run(
        self,
        arg: T.Optional[str] = None,
        debug: bool = False,
    ):
        """
        High level script filter runner.

        :param debug: flag to turn on debug.

        By default, it provides two ways to debug:

        1. Automatically log the python traceback logs to ``~/.alfred-afwf/last-error.txt``
            file.
        2. If python raises any exception, log the last Exception message as an item.
        """
        try:
            self._run(arg=arg, debug=debug)
        except Exception as e:
            if debug:
                log_last_error()
            sf = ScriptFilter()
            item = Item(
                title=f"Error: {e}",
                subtitle=f"Open {str(p_last_error)} to see details",
                icon=Icon.from_image_file(IconFileEnum.error),
                arg=str(p_last_error),
            )
            item._open_last_error_file(path=str(p_last_error))
            sf.items.append(item)
            json.dump(sf.to_script_filter(), sys.stdout)
            sys.stdout.flush()
            exit(1)
        exit(0)



# -*- coding: utf-8 -*-

"""
Run this script to generate ``icon.py`` file.

Do this if you update the ``icons`` folder.
"""

from pathlib_mate import Path

dir_here = Path(__file__).absolute().parent
dir_icons = Path(dir_here, "icons")
p_icon = Path(dir_here, "icon.py")

content = """
# -*- coding: utf-8 -*-

\"\"\"
Enumerate all built-in ICON images.
\"\"\"

import os

dir_icons = os.path.join(os.path.dirname(os.path.abspath(__file__)), "icons")

class IconFileEnum:
""".strip()
lines = content.split("\n")

for p in dir_icons.select_by_ext(".png"):
    name = p.basename.split("-")[0]
    file = p.basename
    line = f"    {name} = os.path.join(dir_icons, \"{file}\")"
    lines.append(line)

lines.append("")

p_icon.write_text("\n".join(lines))



# -*- coding: utf-8 -*-

"""
Enhancement for standard enum library.
"""

import enum
from typing import List


class BetterEnum(enum.Enum):
    """
    Provide additional method for enum.
    """

    @classmethod
    def has_name(cls, name: str) -> bool:
        """
        Check if a name is defined as enum.
        """
        try:
            _ = cls[name]
            return True
        except KeyError:
            return False

    @classmethod
    def has_value(cls, value) -> bool:
        """
        Check if a value is defined as enum.
        """
        try:
            _ = cls(value)
            return True
        except ValueError:
            return False

    @classmethod
    def to_names(cls) -> List[str]:
        """
        Get list of all defined names.
        """
        return [
            i.name
            for i in cls
        ]

    @classmethod
    def to_values(cls) -> list:
        """
        Get list of all defined values.
        """
        return [
            i.value
            for i in cls
        ]


special_char = ["!@#$%^&*()_+-=~`[]{}|,. "]


def snake_case(s) -> str: # pragma: no cover
    """
    convert string to snake case.
    """
    s = s.lower()
    for c in special_char:
        s = s.replace(c, "_")
    return s



# -*- coding: utf-8 -*-

"""
query utilities.
"""

import typing as T
import attr
from attrs_mate import AttrsClass


@attr.define
class Query(AttrsClass):
    """
    Structured query object.

    :param parts: the parts of query string split by delimiter
    :param trimmed_parts: similar to parts, but each part is white space stripped
    """
    parts: T.List[str] = AttrsClass.ib_list_of_str()
    trimmed_parts: T.List[str] = AttrsClass.ib_list_of_str()


SEP = "____"

class QueryParser:
    """
    Utility class that can parse string to query.
    """

    def __init__(
        self,
        delimiter: T.Union[str, T.List[str]] = " ",
    ):
        if isinstance(delimiter, str):
            self.delimiter = [delimiter,]
        else:
            self.delimiter = delimiter

    def parse(self, s: str) -> Query:
        """
        Convert string query to structured query object.
        """
        for sep in self.delimiter:
            s = s.replace(sep, SEP)
        parts = s.split(SEP)
        trimmed_parts = [c.strip() for c in parts if c.strip()]
        return Query(
            parts=parts,
            trimmed_parts=trimmed_parts,
        )



# -*- coding: utf-8 -*-

"""
Frequently used paths.
"""

from pathlib import Path

dir_here = Path(__file__).absolute().parent
dir_lib = Path(dir_here.parent, "lib")

dir_home = Path.home()
dir_afwf = Path(dir_home, ".alfred-afwf")
p_last_error = Path(dir_afwf, "last-error.txt")
p_debug_log = Path(dir_afwf, "debug.txt")



# -*- coding: utf-8 -*-

"""

"""

import attr
from attrs_mate import AttrsClass


@attr.define
class ScriptFilterObject(AttrsClass):
    """
    [CN]

    所有的 ScriptFilter 中出现的 Object 比如 Item, 或是 ScriptFilter 本身都是这个
    类的子类.

    根据 Alfred 的 `官方文档 <https://www.alfredapp.com/help/workflows/inputs/script-filter/json/>`_
    Script Filter 是一个 JSON 的数据结构. 但这个 JSON 数据结构与 Python OOP 的序列化接口
    有差别. OOP 序列化接口是为了能从 JSON 中还原 object. 而 Script Filter 接口是为
    Alfred 服务的. 例如 OOP 中如果一个 attribute 的值为 None, 那么序列化后还是需要保留
    这个值. Script Filter 中如果一个 attribute 被解读为 Omit, 这个 attribute 就彻底
    不应该存在, 不然会出现歧义. 所以 ScriptFilterObject 是一种特殊的对象, 同时提供了
    OOP 序列化接口 和 Script Filter 序列化接口.
    """

    def to_script_filter(self) -> dict:
        """
        Convert object to Alfred Workflow Script Filter friendly dictionary.
        """
        dct = dict()
        for k, v in attr.asdict(self, recurse=False).items():
            if v:
                if isinstance(v, ScriptFilterObject):
                    v1 = v.to_script_filter()
                    if v1:
                        dct[k] = v1
                elif isinstance(v, list):
                    lst = list()
                    for i in v:
                        if isinstance(i, ScriptFilterObject):
                            lst.append(i.to_script_filter())
                        else:
                            lst.append(i)
                    dct[k] = lst
                else:
                    dct[k] = v
        return dct



# -*- coding: utf-8 -*-

from typing import Any, Union, List, Dict

import attr
from attr import validators as vs
from attrs_mate import AttrsClass

from .enumeration import BetterEnum
from .script_filter_object import ScriptFilterObject


@attr.define
class Icon(ScriptFilterObject):
    """ """

    class TypeEnum(BetterEnum):
        fileicon = "fileicon"
        filetype = "filetype"

    path: str = AttrsClass.ib_str()
    type: str = attr.field(
        validator=vs.optional(vs.in_(TypeEnum.to_values())),
        default=None,
    )

    @classmethod
    def from_image_file(cls, path: str) -> "Icon":
        """
        Create an Icon object that using a file on local file system as an icon.
        """
        return cls(path=path)


class VarKeyEnum(BetterEnum):
    open_file = "open_file"
    open_file_path = "open_file_path"
    launch_app_or_file = "launch_app_or_file"
    launch_app_or_file_path = "launch_app_or_file_path"
    reveal_file_in_finder = "reveal_file_in_finder"
    reveal_file_in_finder_path = "reveal_file_in_finder_path"
    browse_in_terminal = "browse_in_terminal"
    browse_in_terminal_path = "browse_in_terminal_path"
    browse_in_alfred = "browse_in_alfred"
    browse_in_alfred_path = "browse_in_alfred_path"
    action_in_alfred = "action_in_alfred"
    file_buffer = "file_buffer"
    default_web_search = "default_web_search"
    open_url = "open_url"
    open_url_arg = "open_url_arg"
    system_command = "system_command"
    itunes_command = "itunes_command"
    run_script = "run_script"
    run_script_arg = "run_script_arg"
    run_ns_apple_script = "run_ns_apple_script"
    run_ns_apple_script_arg = "run_ns_apple_script_arg"
    terminal_command = "terminal_command"
    terminal_command_arg = "terminal_command_arg"
    send_notification = "send_notification"
    send_notification_title = "send_notification_title"
    send_notification_subtitle = "send_notification_subtitle"

    _open_last_error_file = "open_last_error_file"
    _open_last_error_file_path = "open_last_error_file_path"


class VarValueEnum(BetterEnum):
    y = "y"
    n = "n"


@attr.define
class Text(ScriptFilterObject):
    copy: str = AttrsClass.ib_str(default=None)
    largetype: str = AttrsClass.ib_str(default=None)


class ModEnum(BetterEnum):
    cmd = "cmd"
    shift = "shift"
    alt = "alt"
    ctrl = "ctrl"
    fn = "fn"


@attr.define
class Item(ScriptFilterObject):
    """
    Data model for alfred dropdown menu items.

    Ref: https://www.alfredapp.com/help/workflows/inputs/script-filter/json/
    """

    class TypeEnum(BetterEnum):
        file = "file"
        file_skipcheck = "file:skipcheck"

    title: str = AttrsClass.ib_str()
    subtitle: str = AttrsClass.ib_str(default=None)
    arg: str = AttrsClass.ib_str(default=None)
    autocomplete: str = AttrsClass.ib_str(default=None)
    icon: Icon = Icon.ib_nested(default=None)
    valid: bool = AttrsClass.ib_bool(default=True)
    uid: str = AttrsClass.ib_str(default=None)
    match: str = AttrsClass.ib_str(default=None)
    type: str = attr.field(
        validator=vs.optional(vs.in_(TypeEnum.to_values())),
        default=None,
    )
    mods: dict = AttrsClass.ib_dict(default=None)
    action: Union[str, List[str], Dict[str, Any]] = attr.ib(default=None)
    text: Text = Text.ib_nested(default=None)
    quicklookurl: str = AttrsClass.ib_str(default=None)
    variables: dict = AttrsClass.ib_dict(factory=dict)

    # --------------------------------------------------------------------------
    # Set attribute value
    # --------------------------------------------------------------------------
    def set_icon(self, path: str) -> "Item":
        self.icon = Icon.from_image_file(path)
        return self

    def set_modifier(
        self,
        mod: str = ModEnum.cmd.value,
        subtitle: subtitle = None,
        arg: str = None,
        valid: bool = True,
    ) -> "Item":
        """
        Add modifier to item. Modifier allow you to return different title,
        subtitle and argument in Alfred drop down menu item when you hit a
        modifier key.

        :param mode: the modifier key
        """
        if mod not in ModEnum.to_values():
            raise ValueError
        dct = {
            k: v
            for k, v in dict(
                subtitle=subtitle,
                arg=arg,
                valid=valid,
            ).items()
            if v
        }
        if self.mods is None:
            self.mods = dict()
        self.mods[mod] = dct
        return self

    # --------------------------------------------------------------------------
    # Set variables
    # --------------------------------------------------------------------------
    def _open_last_error_file(self, path: str):
        """
        This is a special variable that will open the last error file in the editor.
        It is for internal implementation only, not for public API.
        """
        self.variables[VarKeyEnum._open_last_error_file.value] = VarValueEnum.y.value
        self.variables[VarKeyEnum._open_last_error_file_path.value] = path

    def open_file(self, path: str):
        """
        Set variables to tell subsequence Alfred action to open a file.

        Use the "Utilities -> Conditional" widget and set: if ``{var:open_file}``
        is equal to "y".

        Use the "Actions -> Open File" widget and set: File: ``{var:open_file_path}``
        """
        self.variables[VarKeyEnum.open_file.value] = VarValueEnum.y.value
        self.variables[VarKeyEnum.open_file_path.value] = path

    def launch_app_or_file(self, path: str):
        self.variables[VarKeyEnum.launch_app_or_file.value] = VarValueEnum.y.value
        self.variables[VarKeyEnum.launch_app_or_file_path.value] = path

    def reveal_file_in_finder(self, path: str):
        self.variables[VarKeyEnum.reveal_file_in_finder.value] = VarValueEnum.y.value
        self.variables[VarKeyEnum.reveal_file_in_finder_path.value] = path

    def browse_in_terminal(self, path: str):
        self.variables[VarKeyEnum.browse_in_terminal.value] = VarValueEnum.y.value
        self.variables[VarKeyEnum.browse_in_terminal_path.value] = path

    def browse_in_alfred(self, path: str):
        self.variables[VarKeyEnum.browse_in_alfred.value] = VarValueEnum.y.value
        self.variables[VarKeyEnum.browse_in_alfred_path.value] = path

    def open_url(self, url: str):
        """
        Set variables to tell subsequence Alfred action to open url.

        Use the "Utilities -> Conditional" widget and set: if ``{var:open_url}``
        is equal to "y".

        Use the "Actions -> Open URL" widget and set:

        - File = ``{var:open_url_arg}``
        """
        self.variables[VarKeyEnum.open_url.value] = VarValueEnum.y.value
        self.variables[VarKeyEnum.open_url_arg.value] = url

    def run_script(self, cmd: str):
        """
        Set variables to tell subsequence Alfred action to run script.

        Use the "Utilities -> Conditional" widget and set: if ``{var:run_script}``
        is equal to "y".

        Use the "Actions -> Run Script" widget and set:

        - Language = ``/bin/bash``
        - "with input as {query}"
        - running instances = "Sequentially"
        - Script = ``{query}``
        """
        self.variables[VarKeyEnum.run_script.value] = VarValueEnum.y.value
        self.variables[VarKeyEnum.run_script_arg.value] = cmd

    def terminal_command(self, cmd: str):
        self.variables[VarKeyEnum.terminal_command.value] = VarValueEnum.y.value
        self.variables[VarKeyEnum.terminal_command_arg.value] = cmd

    def send_notification(self, title: str, subtitle: str = ""):
        """
        Set variables to tell subsequence Alfred action to send notification.

        Use the "Utilities -> Conditional" widget and set: if ``{var:send_notification}``
        is equal to "y".

        Use the "Outputs -> Post Notification" widget and set:

        - Title = ``{var:send_notification_title}``
        - Subtitle = ``{var:send_notification_subtitle}``
        """
        self.variables[VarKeyEnum.send_notification.value] = VarValueEnum.y.value
        self.variables[VarKeyEnum.send_notification_title.value] = title
        self.variables[VarKeyEnum.send_notification_subtitle.value] = subtitle



# -*- coding: utf-8 -*-

"""

"""

import attr
from attrs_mate import AttrsClass
from .script_filter import ScriptFilter


@attr.define
class Handler(AttrsClass):
    """
    [CN]

    Handler 主要负责处理 Input Output 逻辑. 在 Alfred Workflow 的编程模型中, Input 是
    搜索框的字符串. Output 则是 Drop Down menu 所需的 Item, 以及一些 Variables, Rerun
    Setting. 这些信息被放在 :class:`~afwf.script_filter.ScriptFilter` 这个 Data
    Container 类中.

    你在 Script Filter 中的 Script 一栏中填写如下信息的时::

        ${path_to_python} main.py '${handler_id} ${query}'

    其行为等效于使用 ``${path_to_python}`` 的 Python 解释器, 调用跟 ``${handler_id}``
    所对应的 :class:`Handler` 类中的 ``handler(query)`` 方法.

    Handler 必须实现两个 methods.

    1. :meth:`~afwf.script_filter.ScriptFilter.lower_level_api`:
    2. :meth:`~afwf.script_filter.ScriptFilter.handler`:
    """
    id = AttrsClass.ib_str(nullable=False)

    def low_level_api(self, **kwargs) -> ScriptFilter:
        """
        [CN]

        一个传统的 Python 函数, 可以自定义参数, 返回的是一个
        :class:`~afwf.script_filter.ScriptFilter` 对象, 里面包含了 Drop Down Menu
        所需的 Item. **最核心的逻辑将在这个函数中被处理**.
        """
        raise NotImplementedError

    def handler(self, query: str) -> ScriptFilter:
        """
        [CN]

        :meth:`~afwf.script_filter.ScriptFilter.low_level_api` 的一个 wrapper
        函数. 只不过接收的参数是一个 query string, 也就是你在 Script Filter Widget
        里面定义的 ``python3 main.py '{handler_id} {query}'`` 中的 query 部分.

        该函数的主要工作是将 query 转化成
        :meth:`~afwf.script_filter.ScriptFilter.low_level_api` 所需的参数.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

from typing import List

import attr
import afwf
import requests
from bs4 import BeautifulSoup

from ..cache import cache


@attr.define
class Handler(afwf.Handler):
    @cache.memoize(expire=60)
    def get_all_python_version(self) -> List[str]:
        invalid_versions = ["patches", "."]
        url = "https://github.com/pyenv/pyenv/tree/master/plugins/python-build/share/python-build"
        res = requests.get(url)
        soup = BeautifulSoup(res.text, "html.parser")
        div = soup.find("div", class_="Box mb-3")
        versions = list()
        for a in div.find_all("a", class_="js-navigation-open Link--primary"):
            version = a.text
            if version not in invalid_versions:
                versions.append(version)
        return versions

    def lower_level_api(self, query: str) -> afwf.ScriptFilter:
        if query == "error":
            raise ValueError("query cannot be 'error'!")
        versions = self.get_all_python_version()
        filtered_versions = [
            version
            for version in versions
            if query in version
        ]
        filtered_versions.sort()
        filtered_versions = filtered_versions[:50]
        sf = afwf.ScriptFilter()
        for version in filtered_versions:
            url = f"https://github.com/pyenv/pyenv/blob/master/plugins/python-build/share/python-build/{version}"
            item = afwf.Item(
                title=version,
                autocomplete=version,
                arg=url,
            )
            item.open_url(url=url)
            sf.items.append(item)
        if len(sf.items) == 0:
            sf.items.append(afwf.Item(
                title=f"{query!r} doesn't match any Python version!",
                subtitle="Open https://github.com/pyenv/pyenv/tree/master/plugins/python-build/share/python-build",
                arg="https://github.com/pyenv/pyenv/tree/master/plugins/python-build/share/python-build",
            ))
        return sf

    def handler(self, query: str) -> afwf.ScriptFilter:
        return self.lower_level_api(query=query)


handler = Handler(id="python_version")



# -*- coding: utf-8 -*-

"""
Alfred Workflow Script Filter power tool.
"""

from ._version import __version__

__short_description__ = "Alfred Workflow Script Filter power tool.."
__license__ = "MIT"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__github_username__ = "MacHu-GWU"

try:
    from .item import (
        Icon,
        Text,
        Item,
        VarKeyEnum,
        VarValueEnum,
        ModEnum,
    )
    from .script_filter import ScriptFilter
    from .handler import Handler
    from .workflow import Workflow
    from .icon import IconFileEnum
except ImportError as e:
    print(e)



# -*- coding: utf-8 -*-

from diskcache import Cache
from pathlib_mate import Path

dir_home = Path.home()
dir_afwf = Path(dir_home, ".alfred-afwf")
dir_cache = Path(dir_afwf, ".cache")

dir_afwf.mkdir_if_not_exists()

cache = Cache(dir_cache.abspath)



# -*- coding: utf-8 -*-

"""
Enumerate all built-in ICON images.
"""

import os

dir_icons = os.path.join(os.path.dirname(os.path.abspath(__file__)), "icons")

class IconFileEnum:
    search = os.path.join(dir_icons, "search-96.png")
    refresh = os.path.join(dir_icons, "refresh-96.png")
    landline = os.path.join(dir_icons, "landline-96.png")
    pause = os.path.join(dir_icons, "pause-96.png")
    calendar = os.path.join(dir_icons, "calendar-96.png")
    start = os.path.join(dir_icons, "start-96.png")
    error = os.path.join(dir_icons, "error-96.png")
    reset = os.path.join(dir_icons, "reset-96.png")
    groupchat = os.path.join(dir_icons, "groupchat-96.png")
    question = os.path.join(dir_icons, "question-96.png")
    chat = os.path.join(dir_icons, "chat-96.png")
    plus = os.path.join(dir_icons, "plus-96.png")
    icons8 = os.path.join(dir_icons, "icons8-message-96.png")
    android = os.path.join(dir_icons, "android-96.png")
    minus = os.path.join(dir_icons, "minus-96.png")
    iphone = os.path.join(dir_icons, "iphone-96.png")
    check = os.path.join(dir_icons, "check-96.png")
    undo = os.path.join(dir_icons, "undo-96.png")
    stop = os.path.join(dir_icons, "stop-96.png")
    redo = os.path.join(dir_icons, "redo-96.png")
    close = os.path.join(dir_icons, "close-96.png")
    todo = os.path.join(dir_icons, "todo-96.png")
    remove = os.path.join(dir_icons, "remove-96.png")
